#include "map.hpp"
#include "dbc.hpp"
#include <vector>
#include <fmt/core.h>
#include <random>

using std::vector, std::pair;
using namespace fmt;

void dump_map(const std::string &msg, Matrix &map) {
  println("----------------- {}", msg);
  for(auto row : map) {
    for(auto col : row) {
      print("{} ", col);
    }
    print("\n");
  }
}

inline void add_neighbors(PairList &neighbors, Matrix &closed, size_t j, size_t i) {
  size_t h = closed.size();
  size_t w = closed[0].size();
  vector<size_t> rows{j - 1, j, j + 1};
  vector<size_t> cols{i - 1, i, i + 1};

  for(auto row : rows) {
    for(auto col : cols) {
      if((0 <= row && row < h) &&
          (0 <= col && col < w) &&
          closed[row][col] == 0)
      {
        closed[row][col] = 1;
        neighbors.push_back({.j=row, .i=col});
      }
    }
  }
}

Map::Map(size_t width, size_t height) : m_limit(1000) {
  m_walls = Matrix(height, MatrixRow(width, 1));
  m_input_map = Matrix(height, MatrixRow(width, 1));
}

void Map::make_paths() {
  size_t h = m_input_map.size();
  size_t w = m_input_map[0].size();

  // Initialize the new array with every pixel at limit distance
  // NOTE: this is normally ones() * limit
  int limit = m_limit == 0 ? h * w : m_limit;
  Matrix new_arr = Matrix(h, MatrixRow(w, limit));
  Matrix closed = m_walls;
  PairList starting_pixels;
  PairList open_pixels;

  // First pass: Add starting pixels and put them in closed
 for(size_t counter = 0; counter < h * w; counter++) {
    size_t i = counter % w;
    size_t j = counter / w;
    if(m_input_map[j][i] == 0) {
      new_arr[j][i] = 0;
      closed[j][i] = 1;
      starting_pixels.push_back({.j=j,.i=i});
    }
  }

  // Second pass: Add border to open
  for(auto sp : starting_pixels) {
    add_neighbors(open_pixels, closed, sp.j, sp.i);
  }

  // Third pass: Iterate filling in the open list
  int counter = 1; // leave this here so it's available below
  for(; counter < limit && !open_pixels.empty(); ++counter) {
    PairList next_open;
    for(auto sp : open_pixels) {
      new_arr[sp.j][sp.i] = counter;
      add_neighbors(next_open, closed, sp.j, sp.i);
    }
    open_pixels = next_open;
  }

  // Last pass: flood last pixels
  for(auto sp : open_pixels) {
    new_arr[sp.j][sp.i] = counter;
  }

  m_paths = new_arr;
}

void Map::make_room(size_t origin_x, size_t origin_y, size_t w, size_t h) {
  dbc::check(origin_x < width(), "x out of bounds");
  dbc::check(origin_y < height(), "x out of bounds");
  dbc::check(w < width(), "x out of bounds");
  dbc::check(h < height(), "x out of bounds");

  for(size_t y = origin_y; y < origin_y + h; ++y) {
    for(size_t x = origin_x; x < origin_x + w; ++x) {
      m_walls[y][x] = 0;
    }
  }
}

struct Partition;

struct Partition {
  size_t x = 0;
  size_t y = 0;
  size_t width = 0;
  size_t height = 0;

  std::vector<Partition> children;
};

inline int make_split(std::mt19937 &gen, Partition &cur, bool horiz) {
  if(horiz) {
    size_t quarter = cur.height / 4;
    // vertical split, pick a random horizontal location
    std::uniform_int_distribution<int> rhoriz(cur.y + quarter, cur.height - quarter);
    return rhoriz(gen);
  } else {
    size_t quarter = cur.width / 4;
    // horizontal split, pick a random vertical location
    std::uniform_int_distribution<int> rvert(cur.x + quarter, cur.width - quarter);
    return rvert(gen);
  }
}

void Map::generate() {
  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> rsplit(0, 1);

  Partition root{
    .x = 1,
    .y = 1,
    .width = width() - 2,
    .height = height() - 2
  };

  Partition &cur = root;
  bool horiz = rsplit(gen);

  for(int i = 0; i < 1; i++) {
    int split = make_split(gen, cur, horiz);

    if(horiz) {
      println("HORIZ split={}, x={}, y={}, w={}, h={}", split, cur.x, cur.y, cur.width, cur.height);

      make_room(cur.x, cur.y, cur.width, split - 1);
      make_room(cur.x, cur.y + split, cur.width, cur.height - split);
    } else {
      println("VERT split={}, x={}, y={}, w={}, h={}", split, cur.x, cur.y, cur.width, cur.height);
      make_room(cur.x, cur.y, split-1, cur.height);
      make_room(cur.x + split, cur.y, cur.width - split, cur.height);
    }
    horiz = !horiz;
  }
}
